use proc_macro2::TokenStream;
use quote::{format_ident, quote, ToTokens};

use crate::model::{MethodModel, ProtocolModel};

impl ToTokens for ProtocolModel {
    fn to_tokens(&self, tokens: &mut TokenStream) {
        tokens.extend(self.generate_message_enum());
        tokens.extend(self.generate_state_traits());
    }
}

impl ProtocolModel {
    fn generate_message_enum(&self) -> TokenStream {
        let variants = self.msg_lookup().msg_iter().map(|msg| msg.msg_name());
        let msg_types = self.msg_lookup().msg_iter().map(|msg| msg.msg_type());
        let enum_name = format_ident!("{}Msgs", self.def().name_def.name);
        quote! {
            pub enum #enum_name {
                #( #variants(#msg_types) ),*
            }
        }
    }

    fn generate_state_traits(&self) -> TokenStream {
        let traits = self
            .states_iter()
            .map(|state| (state.name(), state.methods().values()));
        let mut tokens = TokenStream::new();
        for (trait_ident, methods) in traits {
            let method_tokens = methods.map(|x| x.generate_tokens());
            quote! {
                pub trait #trait_ident {
                    #( #method_tokens )*
                }
            }
            .to_tokens(&mut tokens);
        }
        tokens
    }
}

impl MethodModel {
    /// Generates the tokens for the code which implements this transition.
    fn generate_tokens(&self) -> TokenStream {
        let method_ident = self.name().as_ref();
        let msg_args = self.inputs().iter();
        let output_decls = self.outputs().iter().flat_map(|output| output.decl());
        let output_types = self.outputs().iter().flat_map(|output| output.type_name());
        let future_name = self.future();
        quote! {
            #( #output_decls )*
            type #future_name: ::std::future::Future<Output = Result<( #( #output_types ),* )>>;
            fn #method_ident(self #( , #msg_args )*) -> Self::#future_name;
        }
    }
}
